import re
from typing import List


def remove_comments(string: str) -> str:
    """
    Remove comments from the string.

    Parameters:
        string: str
            The SQL query string from which comments are to be removed.

    Returns:
        str: The SQL query string with comments removed.
    """
    # First group captures quoted strings (double or single)
    # Second group captures comments (/* multi-line */ or -- single-line)
    pattern = r"(\"[^\"]*\"|\'[^\']*\')|(/\*[\s\S]*?\*/|--[^\r\n]*$)"

    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        if match.group(2) is not None:
            return ""  # Remove the comment
        else:
            return match.group(1)  # Keep the quoted string

    return regex.sub(_replacer, string)


def clean_statement(string: str) -> str:
    """
    Remove carriage returns and all whitespace to single spaces.

    Avoid removing whitespace in quoted strings.
    """
    pattern = r"(\"[^\"]*\"|\'[^\']*\'|\`[^\`]*\`)|(\s+)"
    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        if match.group(2) is not None:
            return " "
        return match.group(1)  # captured quoted-string

    return regex.sub(_replacer, string).strip()


def split_sql_statements(sql: str) -> List[str]:
    """
    Splits multiple SQL statements separated by semicolons into a list.

    Parameters:
        sql: str
            A string containing one or more SQL statements.

    Returns:
        List[str]: A list of individual SQL statements.
    """
    statements = []
    buffer: list = []
    in_single_quote = False
    in_double_quote = False
    in_backtick_quote = False

    for char in sql:
        if char == "'" and not in_double_quote and not in_backtick_quote:
            in_single_quote = not in_single_quote
        elif char == '"' and not in_single_quote and not in_backtick_quote:
            in_double_quote = not in_double_quote
        elif char == "`" and not in_single_quote and not in_double_quote:
            in_backtick_quote = not in_backtick_quote
        elif char == ";" and not in_single_quote and not in_double_quote and not in_backtick_quote:
            statements.append("".join(buffer).strip())
            buffer = []
            continue

        buffer.append(char)

    # Append any remaining text
    if buffer:
        statements.append("".join(buffer).strip())

    return [s for s in statements if s != ""]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides a PEP-249 familiar interface for interacting with mabel data
stores, it is not compliant with the standard:
https://www.python.org/dev/peps/pep-0249/
"""

from typing import Iterable
from typing import Optional
from typing import Set

from opteryx.cursor import Cursor
from opteryx.exceptions import PermissionsError
from opteryx.exceptions import ProgrammingError
from opteryx.models import ConnectionContext


class Connection:
    """
    A connection
    """

    def __init__(
        self,
        *,
        user: Optional[str] = None,
        permissions: Optional[Iterable[str]] = None,
        memberships: Optional[Iterable[str]] = None,
        **kwargs,
    ):
        """
        A virtual connection to the Opteryx query engine.
        """
        self._kwargs = kwargs

        if memberships and not all(isinstance(v, str) for v in memberships):
            raise ProgrammingError("Invalid memberships provided to Connection")
        if permissions and not all(isinstance(v, str) for v in permissions):
            raise ProgrammingError("Invalid permissions provided to Connection")
        if user and not isinstance(user, str):
            raise ProgrammingError("Invalid user provided to Connection")

        self.context = ConnectionContext(user=user, memberships=memberships)

        # check the permissions we've been given are valid permissions
        self.permissions = self.validate_permissions(permissions)

    def cursor(self) -> Cursor:
        """return a cursor object"""
        return Cursor(self)

    def close(self) -> None:
        """Exists for interface compatibility only."""
        pass

    def commit(self) -> None:
        """Exists for interface compatibility only."""
        pass

    def rollback(self):
        """Exists for interface compatibility only. Raises AttributeError."""
        # return AttributeError as per https://peps.python.org/pep-0249/#id48
        raise AttributeError("Opteryx does not support transactions.")

    @property
    def history(self):
        """Returns a list of queries previously run on this connection"""
        return self.context.history

    @staticmethod
    def validate_permissions(permissions: Optional[Iterable[str]] = None) -> Set[str]:
        """
        This is checking the validity of the permissions provided, not that the user has the
        right permissions.

        Parameters:
            permissions: Optional[Iterable[str]], optional
                Permissions for the connection.
        Returns:
            Validated set of permissions.
        """
        from opteryx.constants.permissions import PERMISSIONS

        if permissions is None:
            permissions = PERMISSIONS
        permissions = set(permissions)

        invalid_permissions = permissions.difference(PERMISSIONS)
        if len(invalid_permissions) > 0:
            raise PermissionsError(f"Invalid permissions presented - {invalid_permissions}")
        return permissions



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Async Scanner Node

This is the SQL Query Execution Plan Node responsible for the reading of data.

It wraps different internal readers (e.g. GCP Blob reader, SQL Reader),
normalizes the data into the format for internal processing.
"""

import asyncio
import queue
import threading
import time
from dataclasses import dataclass
from typing import Generator

import aiohttp
import pyarrow
import pyarrow.parquet
from orso.schema import RelationSchema

from opteryx import config
from opteryx.operators.base_plan_node import BasePlanDataObject
from opteryx.operators.read_node import ReaderNode
from opteryx.shared import AsyncMemoryPool
from opteryx.shared import MemoryPool
from opteryx.utils.file_decoders import get_decoder

CONCURRENT_READS = config.CONCURRENT_READS
MAX_READ_BUFFER_CAPACITY = config.MAX_READ_BUFFER_CAPACITY


def normalize_morsel(schema: RelationSchema, morsel: pyarrow.Table) -> pyarrow.Table:
    if len(schema.columns) == 0:
        one_column = pyarrow.array([1] * morsel.num_rows, type=pyarrow.int8())
        morsel = morsel.append_column("*", one_column)
        return morsel.select(["*"])

    # rename columns for internal use
    target_column_names = []
    # columns in the data but not in the schema, droppable
    droppable_columns = []

    # find which columns to drop and which columns we already have
    for i, column in enumerate(morsel.column_names):
        column_name = schema.find_column(column)
        if column_name is None:
            droppable_columns.append(i)
        else:
            target_column_names.append(str(column_name))

    # remove from the end otherwise we'll remove the wrong columns after the first one
    droppable_columns.reverse()
    for droppable in droppable_columns:
        morsel = morsel.remove_column(droppable)

    # remane columns to the internal names (identities)
    morsel = morsel.rename_columns(target_column_names)

    # add columns we don't have
    for column in schema.columns:
        if column.identity not in target_column_names:
            null_column = pyarrow.array([None] * morsel.num_rows)
            morsel = morsel.append_column(column.identity, null_column)

    # ensure the columns are in the right order
    return morsel.select([col.identity for col in schema.columns])


async def fetch_data(blob_names, pool, reader, reply_queue, statistics):
    semaphore = asyncio.Semaphore(CONCURRENT_READS)
    session = aiohttp.ClientSession()

    async def fetch_and_process(blob_name):
        async with semaphore:
            start_per_blob = time.monotonic_ns()
            reference = await reader(
                blob_name=blob_name, pool=pool, session=session, statistics=statistics
            )
            reply_queue.put((blob_name, reference))  # Put data onto the queue
            statistics.time_reading_blobs += time.monotonic_ns() - start_per_blob

    tasks = (fetch_and_process(blob) for blob in blob_names)

    await asyncio.gather(*tasks)
    reply_queue.put(None)
    await session.close()


@dataclass
class AsyncReaderDataObject(BasePlanDataObject):
    pass


class AsyncReaderNode(ReaderNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pool = MemoryPool(MAX_READ_BUFFER_CAPACITY, f"ReadBuffer <{self.parameters['alias']}>")

        self.do = AsyncReaderDataObject()

    @classmethod
    def from_dict(cls, dic: dict) -> "AsyncReaderNode":  # pragma: no cover
        raise NotImplementedError()

    def execute(self) -> Generator:
        """Perform this step, time how long is spent doing work"""
        orso_schema = self.parameters["schema"]
        reader = self.parameters["connector"]

        blob_names = reader.partition_scheme.get_blobs_in_partition(
            start_date=reader.start_date,
            end_date=reader.end_date,
            blob_list_getter=reader.get_list_of_blob_names,
            prefix=reader.dataset,
        )

        data_queue: queue.Queue = queue.Queue()

        loop = asyncio.new_event_loop()
        threading.Thread(
            target=lambda: loop.run_until_complete(
                fetch_data(
                    blob_names,
                    AsyncMemoryPool(self.pool),
                    reader.async_read_blob,
                    data_queue,
                    self.statistics,
                )
            ),
            daemon=True,
        ).start()

        orso_schema_cols = []
        for col in orso_schema.columns:
            if col.identity in [c.schema_column.identity for c in self.columns]:
                orso_schema_cols.append(col)
        orso_schema.columns = orso_schema_cols

        morsel = None
        arrow_schema = None

        while True:
            try:
                # Attempt to get an item with a timeout.
                item = data_queue.get(timeout=0.1)
            except queue.Empty:
                # Increment stall count if the queue is empty.
                self.statistics.stalls_reading_from_read_buffer += 1
                continue  # Skip the rest of the loop and try to get an item again.

            if item is None:
                # Break out of the loop if the item is None, indicating a termination condition.
                break

            blob_name, reference = item

            decoder = get_decoder(blob_name)
            # This pool is being used by async processes in another thread, using
            # zero copy versions occassionally results in data getting corrupted
            blob_bytes = self.pool.read_and_release(reference, zero_copy=False)
            try:
                start = time.monotonic_ns()
                # the sync readers include the decode time as part of the read time
                decoded = decoder(blob_bytes, projection=self.columns, selection=self.predicates)
                self.statistics.time_reading_blobs += time.monotonic_ns() - start
                num_rows, _, morsel = decoded
                self.statistics.rows_seen += num_rows

                morsel = normalize_morsel(orso_schema, morsel)
                if arrow_schema:
                    morsel = morsel.cast(arrow_schema)
                else:
                    arrow_schema = morsel.schema

                self.statistics.blobs_read += 1
                self.statistics.rows_read += morsel.num_rows
                self.statistics.bytes_processed += morsel.nbytes

                yield morsel
            except Exception as err:
                self.statistics.add_message(f"failed to read {blob_name}")
                self.statistics.failed_reads += 1
                print(f"[READER] Cannot read blob {blob_name} due to {err}")
                raise err

        if morsel:
            self.statistics.columns_read += morsel.num_columns



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import time
from enum import Enum
from enum import auto
from functools import wraps
from itertools import chain
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union
from uuid import uuid4

import pyarrow
from orso import DataFrame
from orso import converters
from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx import config
from opteryx import utils
from opteryx.constants import QueryStatus
from opteryx.constants import ResultType
from opteryx.exceptions import InconsistentSchemaError
from opteryx.exceptions import InvalidCursorStateError
from opteryx.exceptions import MissingSqlStatement
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.models import QueryStatistics
from opteryx.shared.rolling_log import RollingLog
from opteryx.utils import sql

PROFILE_LOCATION = config.PROFILE_LOCATION
QUERY_LOG_LOCATION = config.QUERY_LOG_LOCATION
QUERY_LOG_SIZE = config.QUERY_LOG_SIZE


ROLLING_LOG = None
if QUERY_LOG_LOCATION:
    ROLLING_LOG = RollingLog(QUERY_LOG_LOCATION, max_entries=QUERY_LOG_SIZE)


class CursorState(Enum):
    INITIALIZED = auto()
    EXECUTED = auto()
    CLOSED = auto()


def require_state(required_state):
    """
    Decorator to enforce a required state before a Cursor method is called.

    The decorator takes a required_state parameter which is the state the Cursor
    must be in before the decorated method can be called. If the state condition
    is not met, an InvalidCursorStateError is raised.

    Parameters:
        required_state: The state that the cursor must be in to execute the method.

    Returns:
        A wrapper function that checks the state and calls the original function.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            if obj._state != required_state:
                raise InvalidCursorStateError(f"Cursor must be in {required_state} state.")
            return func(obj, *args, **kwargs)

        return wrapper

    return decorator


def transition_to(new_state):
    """
    Decorator to transition the Cursor to a new state after a method call.

    The decorator takes a new_state parameter which is the state the Cursor
    will transition to after the decorated method is called.

    Parameters:
        new_state: The new state to transition to after the method is called.

    Returns:
        A wrapper function that executes the original function and then updates the state.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            # Execute the original method.
            result = func(obj, *args, **kwargs)
            # Transition to the new state.
            obj._state = new_state
            return result

        return wrapper

    return decorator


class Cursor(DataFrame):
    """
    This class inherits from the orso DataFrame library to provide features such as fetch.

    This class includes custom decorators @require_state and @transition_to for state management.
    """

    def __init__(self, connection):
        """
        Initializes the Cursor object, setting the initial state and binding the connection.

        Parameters:
            connection: Connection object
                The database connection object.
        """
        self.arraysize = 1
        self._connection = connection
        self._query_planner = None
        self._collected_stats = None
        self._plan = None
        self._qid = str(uuid4())
        self._statistics = QueryStatistics(self._qid)
        self._state = CursorState.INITIALIZED
        self._query_status = QueryStatus._UNDEFINED
        self._result_type = ResultType._UNDEFINED
        self._rowcount = None
        DataFrame.__init__(self, rows=[], schema=[])

    @property
    def id(self) -> str:
        """The unique internal reference for this query.

        Returns:
            The unique query identifier as a string.
        """
        return self._qid

    def _inner_execute(self, operation: str, params: Union[Iterable, Dict, None] = None) -> Any:
        """
        Executes a single SQL operation within the current cursor.

        Parameters:
            operation: str
                SQL operation to be executed.
            params: Iterable/Dictionary, optional
                Parameters for the SQL operation, defaults to None.
        Returns:
            Results of the query execution.
        """

        from opteryx.planner import query_planner

        if not operation:  # pragma: no cover
            raise MissingSqlStatement("SQL provided was empty.")

        self._connection.context.history.append((operation, True, datetime.datetime.utcnow()))

        start = time.time_ns()
        plans = query_planner(
            operation=operation,
            parameters=params,
            connection=self._connection,
            qid=self.id,
            statistics=self._statistics,
        )

        try:
            start = time.time_ns()
            first_item = next(plans)
            self._statistics.time_planning += time.time_ns() - start
        except RuntimeError:  # pragma: no cover
            raise MissingSqlStatement(
                "SQL statement provided had no executable part, this may mean the statement was commented out."
            )

        plans = chain([first_item], plans)

        if ROLLING_LOG:
            ROLLING_LOG.append(operation)

        results = None
        start = time.time_ns()
        for plan in plans:
            self._statistics.time_planning += time.time_ns() - start
            results = plan.execute()
            start = time.time_ns()

        if results is not None:
            # we can't update tuples directly
            self._connection.context.history[-1] = tuple(
                True if i == 1 else value
                for i, value in enumerate(self._connection.context.history[-1])
            )
            return results

    def _execute_statements(self, operation, params: Optional[Iterable] = None):
        """
        Executes one or more SQL statements, properly handling comments, cleaning, and splitting.

        Parameters:
            operation: str
                SQL operation(s) to be executed.
            params: Iterable, optional
                Parameters for the SQL operation(s), defaults to None.

        Returns:
            Results of the query execution, if any.
        """
        self._statistics.start_time = time.time_ns()

        operation = sql.remove_comments(operation)
        operation = sql.clean_statement(operation)
        statements = sql.split_sql_statements(operation)

        if len(statements) == 0:
            raise MissingSqlStatement("No statement found")

        if len(statements) > 1 and params is not None and not isinstance(params, dict) and params:
            raise UnsupportedSyntaxError(
                "Batched queries cannot be parameterized with parameter lists, use named parameters."
            )

        results = None
        for index, statement in enumerate(statements):
            results = self._inner_execute(statement, params)
            if index < len(statements) - 1:
                for _ in results:
                    pass

        # we only return the last result set
        return results

    @require_state(CursorState.INITIALIZED)
    @transition_to(CursorState.EXECUTED)
    def execute(self, operation: str, params: Optional[Iterable] = None):
        """
        Executes the provided SQL operation, converting results to internal DataFrame format.

        Parameters:
            operation: str
                SQL operation to be executed.
            params: Iterable, optional
                Parameters for the SQL operation, defaults to None.
        """
        if hasattr(operation, "decode"):
            operation = operation.decode()
        results = self._execute_statements(operation, params)
        if results is not None:
            result_data, self._result_type = next(results, (ResultType._UNDEFINED, None))
            if self._result_type == ResultType.NON_TABULAR:
                import orso

                meta_dataframe = orso.DataFrame(
                    rows=[(result_data.record_count,)],  # type: ignore
                    schema=RelationSchema(
                        name="table",
                        columns=[FlatColumn(name="rows_affected", type=OrsoTypes.INTEGER)],
                    ),
                )  # type: ignore
                self._rows = meta_dataframe._rows
                self._schema = meta_dataframe._schema

                self._rowcount = result_data.record_count  # type: ignore
                self._query_status = result_data.status  # type: ignore
            elif self._result_type == ResultType.TABULAR:
                self._rows, self._schema = converters.from_arrow(result_data)
                self._cursor = iter(self._rows)
                self._query_status = QueryStatus.SQL_SUCCESS
            else:  # pragma: no cover
                self._query_status = QueryStatus.SQL_FAILURE

    @property
    def result_type(self) -> ResultType:
        return self._result_type

    @property
    def query_status(self) -> QueryStatus:
        return self._query_status

    @property
    def rowcount(self) -> int:
        if self._result_type == ResultType.TABULAR:
            return super().rowcount
        if self._result_type == ResultType.NON_TABULAR:
            return self._rowcount
        raise InvalidCursorStateError("Cursor not in valid state to return a row count.")

    @require_state(CursorState.INITIALIZED)
    @transition_to(CursorState.EXECUTED)
    def execute_to_arrow(
        self,
        operation: str,
        params: Optional[Iterable] = None,
        limit: Optional[int] = None,
    ) -> pyarrow.Table:
        """
        Executes the SQL operation, bypassing conversion to Orso and returning directly in Arrow format.

        Parameters:
            operation: str
                SQL operation to be executed.
            params: Iterable, optional
                Parameters for the SQL operation, defaults to None.
            limit: int, optional
                Limit on the number of records to return, defaults to all records.

        Returns:
            The query results in Arrow table format.
        """
        if hasattr(operation, "decode"):
            operation = operation.decode()
        results = self._execute_statements(operation, params)
        if results is not None:
            result_data, self._result_type = next(results, (ResultType._UNDEFINED, None))
            if limit is not None:
                result_data = utils.arrow.limit_records(result_data, limit)  # type: ignore
        try:
            return pyarrow.concat_tables(result_data, promote_options="permissive")
        except (
            pyarrow.ArrowInvalid,
            pyarrow.ArrowTypeError,
        ) as err:  # pragma: no cover
            # DEBUG: log (err)
            if "struct" in str(err):
                raise InconsistentSchemaError(
                    "Unable to resolve different schemas, most likely related to a STRUCT column."
                ) from err
            raise InconsistentSchemaError(
                "Unable to resolve different schemas, this may be due to uncoercible column types."
            ) from err

    @property
    def stats(self) -> Dict[str, Any]:
        """
        Gets the execution statistics.

        Returns:
            Dictionary containing query execution statistics.
        """
        if self._statistics.end_time == 0:  # pragma: no cover
            self._statistics.end_time = time.time_ns()
        return self._statistics.as_dict()

    @property
    def messages(self) -> List[str]:
        """
        Gets the list of run-time warnings.

        Returns:
            List of warnings generated during query execution.
        """
        return self._statistics.messages

    @require_state(CursorState.EXECUTED)
    @transition_to(CursorState.CLOSED)
    def close(self):
        """
        Closes the cursor, releasing any resources and closing the associated connection.
        """
        self._connection.close()

    def __repr__(self):  # pragma: no cover
        """
        Override the Orso repr

        In notebooks we should return a table
        """
        in_a_notebook = False
        try:
            from IPython import get_ipython

            in_a_notebook = get_ipython() is not None
            if not in_a_notebook:
                return f"<opteryx.Cursor {self._state}>"
        except Exception:  # nosec
            pass
        return str(self)

    def __bool__(self):
        """
        Truthy if executed, Falsy if not executed or error
        """
        return self._state == CursorState.EXECUTED



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import concurrent.futures
import datetime
import os
from typing import Callable
from typing import List
from typing import Optional

from opteryx.exceptions import DataError
from opteryx.managers.schemes import BasePartitionScheme
from opteryx.utils.dates import date_range
from opteryx.utils.file_decoders import DATA_EXTENSIONS

OS_SEP = os.sep


class UnsupportedSegementationError(DataError):
    """Exception raised for unsupported segmentations."""

    def __init__(self, dataset: str, segment: str):
        self.dataset = dataset
        self.segment = segment
        message = f"'{dataset}' contains unsupported segmentation (found `{segment}`), only 'by_hour' segments are supported."
        super().__init__(message)


def extract_prefix(path, prefix):
    start_index = path.find(prefix)
    end_index = path.find(OS_SEP, start_index) if start_index != -1 else -1
    return path[start_index:end_index] if end_index != -1 else None


def is_complete_and_not_invalid(blobs, as_at):
    # Directly initialize flags to False.
    complete = False
    ignore = False

    # Search suffix strings are constant for each call, so define them once.
    complete_suffix = f"{as_at}{OS_SEP}frame.complete"
    invalid_suffix = f"{as_at}{OS_SEP}frame.ignore"

    # Iterate over blobs once, checking conditions.
    for blob in blobs:
        if complete_suffix in blob:
            complete = True
            if complete and ignore:
                break
        elif invalid_suffix in blob:
            ignore = True
            if complete and ignore:
                break

    return complete and not ignore


class MabelPartitionScheme(BasePartitionScheme):
    """
    Handle reading data using the Mabel partition scheme.
    """

    def get_blobs_in_partition(
        self,
        *,
        blob_list_getter: Callable,
        prefix: str,
        start_date: Optional[datetime.datetime],
        end_date: Optional[datetime.datetime],
    ) -> List[str]:
        """filter the blobs acording to the chosen scheme"""

        midnight = datetime.datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
        by_label = f"{OS_SEP}by_"
        as_at_label = f"{OS_SEP}as_at"

        def process_as_ats(blobs, as_at_label, control_blobs):
            as_ats = sorted(
                {extract_prefix(blob, "as_at_") for blob in blobs if as_at_label in blob}
            )
            if len(as_ats) == 0:
                return blobs
            valid_blobs = []
            while as_ats:
                as_at = as_ats.pop()
                if as_at is None:
                    continue
                if is_complete_and_not_invalid(control_blobs, as_at):
                    valid_blobs = [blob for blob in blobs if as_at in blob]
                    break
                else:
                    blobs = [blob for blob in blobs if as_at not in blob]
            return valid_blobs

        def _inner(*, date, start, end):
            date_path = f"{prefix}{OS_SEP}year_{date.year:04d}{OS_SEP}month_{date.month:02d}{OS_SEP}day_{date.day:02d}"
            by_label = f"{OS_SEP}by_"
            as_at_label = f"{OS_SEP}as_at"

            # Call your method to get the list of blob names
            blob_names = blob_list_getter(prefix=date_path)
            if len(blob_names) == 0:
                return []

            control_blobs: List[str] = []
            data_blobs: List[str] = []

            for blob in blob_names:
                ext = os.path.splitext(blob)[1]
                if ext not in DATA_EXTENSIONS:
                    control_blobs.append(blob)
                else:
                    data_blobs.append(blob)
                    if by_label in blob:
                        segment = extract_prefix(blob, "by_")
                        if segment != "by_hour":
                            raise UnsupportedSegementationError(dataset=prefix, segment=segment)

            if any(f"{OS_SEP}by_hour{OS_SEP}" in blob_name for blob_name in data_blobs):
                start = min(start, date)
                end = max(end, date)

                selected_blobs = []

                for hour in date_range(start, end, "1h"):
                    hour_label = f"{OS_SEP}by_hour{OS_SEP}hour={hour.hour:02d}/"
                    # Filter for the specific hour, if hour folders exist
                    if any(hour_label in blob_name for blob_name in data_blobs):
                        hours_blobs = [
                            blob_name for blob_name in data_blobs if hour_label in blob_name
                        ]
                        selected_blobs.extend(
                            process_as_ats(hours_blobs, as_at_label, control_blobs)
                        )
            else:
                selected_blobs = process_as_ats(data_blobs, "as_at_", control_blobs)

            return selected_blobs

        start_date = start_date or midnight
        end_date = end_date or midnight.replace(hour=23, minute=59)

        found = set()

        # Use a ThreadPoolExecutor to parallelize fetching blobs for each hour
        with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
            # Prepare a list of future tasks
            futures = [
                executor.submit(_inner, **{"date": date, "start": start_date, "end": end_date})
                for date in date_range(start_date, end_date, "1d")
            ]
            # Wait for all futures to complete and collect results
            for future in concurrent.futures.as_completed(futures):
                found.update(future.result())

        return sorted(found)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Join Node

This is a SQL Query Execution Plan Node.

This handles most of the join types as a wrapper for pyarrow's JOIN functions,
except for INNER JOIN and CROSS JOIN. PyArrow has a very good set of JOIN
implementations which use here, we don't use the INNER JOIN because PyArrow
has limitations on the column types allowed in the relations, so we use our
own INNER JOIN algo.
"""

from typing import Generator

import pyarrow

from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.operators import OperatorType


class JoinNode(BasePlanNode):
    operator_type = OperatorType.PASSTHRU

    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._join_type = config["type"]
        self._on = config.get("on")
        self._using = config.get("using")

        self._left_columns = config.get("left_columns")
        self._left_relation = config.get("left_relation_names")

        self._right_columns = config.get("right_columns")
        self._right_relation = config.get("right_relation_names")

    @classmethod
    def from_json(cls, json_obj: str) -> "BasePlanNode":  # pragma: no cover
        raise NotImplementedError()

    @property
    def name(self):  # pragma: no cover
        return f"{self._join_type} Join"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Generator:
        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        right_table = pyarrow.concat_tables(right_node.execute(), promote_options="none")

        for morsel in left_node.execute():
            # in place until #1295 resolved
            if self._right_columns[0] not in morsel.column_names:
                self._right_columns, self._left_columns = (
                    self._left_columns,
                    self._right_columns,
                )

            try:
                # do the join
                new_morsel = morsel.join(
                    right_table,
                    keys=self._right_columns,
                    right_keys=self._left_columns,
                    join_type=self._join_type,
                    coalesce_keys=self._using is not None,
                )
            except pyarrow.ArrowInvalid as err:  # pragma: no cover
                last_token = str(err).split(" ")[-1]
                column = None
                for col in left_node.columns:
                    if last_token == col.identity:
                        column = col.name
                        break
                for col in right_node.columns:
                    if last_token == col.identity:
                        column = col.name
                        break
                if column:
                    raise UnsupportedSyntaxError(
                        f"Unable to JOIN with unsupported column types in table, '{column}'."
                    ) from err
                raise UnsupportedSyntaxError(
                    "Unable to JOIN with unsupported column types in table."
                ) from err

            yield new_morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

import numpy
import pyarrow
from pyarrow import compute

from opteryx.exceptions import InvalidFunctionParameterError
from opteryx.exceptions import SqlError
from opteryx.utils.dates import parse_iso


def date_part(part, arr):
    """
    Also the EXTRACT function - we extract a given part from an array of dates
    """
    j2000_scalar = numpy.array([numpy.datetime64("2000-01-01T12:00:00", "us")])
    extractors = {
        "nanosecond": compute.nanosecond,
        "nanoseconds": compute.nanosecond,
        "microsecond": compute.microsecond,
        "microseconds": compute.microsecond,
        "millisecond": compute.millisecond,
        "milliseconds": compute.millisecond,
        "second": compute.second,
        "minute": compute.minute,
        "hour": compute.hour,
        "time": lambda x: compute.cast(x, "time64[us]"),
        "day": compute.day,
        "dayofweek": compute.day_of_week,
        "dow": compute.day_of_week,
        "date": lambda x: compute.cast(x, "date32"),
        "week": compute.week,
        "isoweek": compute.iso_week,
        "month": compute.month,
        "quarter": compute.quarter,
        "dayofyear": compute.day_of_year,
        "doy": compute.day_of_year,
        "year": compute.year,
        "isoyear": compute.iso_year,
        "decade": lambda x: compute.divide(compute.year(x), 10),
        "century": lambda x: compute.add(compute.divide(compute.year(x), 100), 1),
        "epoch": lambda x: compute.divide(compute.cast(x, "int64"), 1000000.00),
        "julian": lambda x: compute.add(
            compute.divide(compute.milliseconds_between(x, j2000_scalar), 86400000.0),
            2451545.0,
        ),
        # ** supported by parser but not by pyarrow
        # isodow
        # millenium
        # millennium
        # timezone
        # time
        # timezonehour
        # timezoneminute
    }

    # if we get a date literal
    if not hasattr(arr, "__iter__"):
        arr = numpy.array([arr])

    part = part[0].lower()  # [#325]
    if part in extractors:
        return extractors[part](arr)

    from opteryx.utils import suggest_alternative

    alt = suggest_alternative(part, list(extractors.keys()))
    if not alt:
        raise InvalidFunctionParameterError(
            f"Date part `{part}` unsupported for EXTRACT."
        )  # pragma: no cover
    raise InvalidFunctionParameterError(
        f"Date part `{part}` unsupported for EXTRACT. Did you mean '{alt}'?"
    )


def date_diff(part, start, end):
    """calculate the difference between timestamps"""

    extractors = {
        "days": compute.days_between,
        "hours": compute.hours_between,
        "microseconds": compute.microseconds_between,
        "minutes": compute.minutes_between,
        "months": compute.month_interval_between,  # this one doesn't work
        "quarters": compute.quarters_between,
        "seconds": compute.seconds_between,
        "weeks": compute.weeks_between,
        "years": compute.years_between,
    }

    # if we get date literals - this will never run due to [#325]
    if isinstance(start, str):  # pragma: no cover
        if not isinstance(end, (str, datetime.datetime)):
            start = pyarrow.array([parse_iso(start)] * len(end), type=pyarrow.timestamp("us"))
        else:
            start = pyarrow.array([parse_iso(start)], type=pyarrow.timestamp("us"))
    if isinstance(end, str):  # pragma: no cover
        if not isinstance(start, (str, datetime.datetime)):
            end = pyarrow.array([parse_iso(end)] * len(start), type=pyarrow.timestamp("us"))
        else:
            end = pyarrow.array([parse_iso(end)], type=pyarrow.timestamp("us"))

    TARGET_DATE_TYPE: str = "datetime64[us]"
    # cast to the desired type
    if hasattr(start, "dtype") and start.dtype != TARGET_DATE_TYPE:
        start = start.astype(TARGET_DATE_TYPE)
    if hasattr(end, "dtype") and end.dtype != TARGET_DATE_TYPE:
        end = end.astype(TARGET_DATE_TYPE)

    part = part[0].lower()  # [#325]
    if part[-1] != "s":
        part += "s"
    if part in extractors:
        diff = extractors[part](start, end)
        if not hasattr(diff, "__iter__"):
            diff = numpy.array([diff])
        return [i.as_py() for i in diff]

    raise SqlError(f"Date part '{part}' unsupported for DATEDIFF")


def time_diff(time1, time2):
    return date_diff(["hours"], time1, time2)


def date_format(dates, pattern):  # [#325]
    pattern = pattern[0]
    return [None if d is None else d.strftime(pattern) for d in dates.tolist()]


def date_floor(dates, magnitude, units):  # [#325]
    return compute.floor_temporal(dates, magnitude[0], units[0])


def from_unixtimestamp(values):
    return [datetime.datetime.fromtimestamp(i) for i in values]


def unixtime(*args):
    if isinstance(args[0], int):
        now = datetime.datetime.utcnow().timestamp()
        return numpy.full(args[0], now, numpy.int64)
    return [numpy.nan if d != d else d.astype(numpy.int64) for d in args[0]]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The BaseConnector provides a common interface for all storage connectors.
"""

from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

import pyarrow
from orso.schema import RelationSchema

from opteryx.models import QueryStatistics

MIN_CHUNK_SIZE: int = 500
INITIAL_CHUNK_SIZE: int = 500
DEFAULT_MORSEL_SIZE: int = 16 * 1024 * 1024


class BaseConnector:
    @property
    def __mode__(self):  # pragma: no cover
        raise NotImplementedError("__mode__ not defined")

    @property
    def interal_only(self):
        return False

    def __init__(
        self,
        *,
        dataset: str = None,
        config: Dict[str, Any] = None,
        statistics: QueryStatistics,
        **kwargs,
    ) -> None:
        """
        Initialize the base connector with configuration.

        Args:
            dataset: The name of the dataset to read.
            config: Configuration information specific to the connector.
        """
        if config is None:
            self.config = {}
        else:
            self.config = config.copy()
        self.dataset = dataset
        self.chunk_size = INITIAL_CHUNK_SIZE
        self.schema = None
        self.statistics = statistics
        self.pushed_predicates: list = []

    def get_dataset_schema(self) -> RelationSchema:  # pragma: no cover
        """
        Retrieve the schema of a dataset.

        Returns:
            A RelationSchema representing the schema of the dataset.
        """
        raise NotImplementedError("Subclasses must implement get_dataset_schema method.")

    def read_dataset(self, **kwargs) -> Iterable:  # pragma: no cover
        """
        Read a dataset and return a reader object.

        Args:
            dataset_name: Name of the dataset.

        Returns:
            A reader object for iterating over the dataset.
        """
        raise NotImplementedError("Subclasses must implement read_dataset method.")

    def read_schema_from_metastore(self):
        # to be implemented
        return None

    def chunk_dictset(
        self,
        dictset: Iterable[dict],
        columns: Optional[list] = None,
        morsel_size: int = DEFAULT_MORSEL_SIZE,
        initial_chunk_size: int = INITIAL_CHUNK_SIZE,
    ) -> pyarrow.Table:
        chunk = []
        self.chunk_size = initial_chunk_size  # we reset each time
        morsel = None

        for index, record in enumerate(dictset):
            _id = record.pop("_id", None)
            # column selection
            if columns:
                record = {k.source_column: record.get(k.source_column) for k in columns}
            record["id"] = None if _id is None else str(_id)

            chunk.append(record)

            if index == self.chunk_size - 1:
                morsel = pyarrow.Table.from_pylist(chunk)
                # Estimate the number of records to fill the morsel size
                if morsel.nbytes > 0:
                    self.chunk_size = int(morsel_size // (morsel.nbytes / self.chunk_size))
                yield morsel
                chunk = []
            elif (index > self.chunk_size - 1) and (index - self.chunk_size) % self.chunk_size == 0:
                morsel = pyarrow.Table.from_pylist(chunk)
                yield morsel
                chunk = []

        if chunk:
            morsel = pyarrow.Table.from_pylist(chunk)
            yield morsel


class DatasetReader:
    def __init__(self, dataset_name: str, config: Optional[Dict[str, Any]] = None) -> None:
        """
        Initialize the reader with configuration.

        Args:
            config: Configuration information specific to the reader.
        """
        self.dataset_name = dataset_name
        self.config = config

    def __iter__(self) -> "DatasetReader":
        """
        Make the reader object iterable.
        """
        return self

    def __next__(self) -> pyarrow.Table:  # pragma: no cover
        """
        Read the next chunk or morsel from the dataset.

        Returns:
            A pyarrow Table representing a chunk or morsel of the dataset.
            raises StopIteration if the dataset is exhausted.
        """
        raise NotImplementedError("Subclasses must implement __next__ method.")

    def close(self) -> None:
        """
        Close the reader and release any resources.
        """
        pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Originally from https://github.com/kelsin/mysql-mimic
"""

# pylint: disable=C0103
from __future__ import annotations

from enum import IntEnum


class CharacterSet(IntEnum):
    big5 = 1
    dec8 = 3
    cp850 = 4
    hp8 = 6
    koi8r = 7
    latin1 = 8
    latin2 = 9
    swe7 = 10
    ascii = 11
    ujis = 12
    sjis = 13
    hebrew = 16
    tis620 = 18
    euckr = 19
    koi8u = 22
    gb2312 = 24
    greek = 25
    cp1250 = 26
    gbk = 28
    latin5 = 30
    armscii8 = 32
    utf8 = 33
    ucs2 = 35
    cp866 = 36
    keybcs2 = 37
    macce = 38
    macroman = 39
    cp852 = 40
    latin7 = 41
    cp1251 = 51
    utf16 = 54
    utf16le = 56
    cp1256 = 57
    cp1257 = 59
    utf32 = 60
    binary = 63
    geostd8 = 92
    cp932 = 95
    eucjpms = 97
    gb18030 = 248
    utf8mb4 = 255

    @property
    def codec(self) -> str:
        if self.name == "utf8mb4":
            return "utf8"
        return self.name

    @property
    def default_collation(self) -> Collation:
        return DEFAULT_COLLATIONS[self]

    def decode(self, b: bytes) -> str:
        return b.decode(self.codec)

    def encode(self, s: str) -> bytes:
        return s.encode(self.codec)


class Collation(IntEnum):
    big5_chinese_ci = 1
    latin2_czech_cs = 2
    dec8_swedish_ci = 3
    cp850_general_ci = 4
    latin1_german1_ci = 5
    hp8_english_ci = 6
    koi8r_general_ci = 7
    latin1_swedish_ci = 8
    latin2_general_ci = 9
    swe7_swedish_ci = 10
    ascii_general_ci = 11
    ujis_japanese_ci = 12
    sjis_japanese_ci = 13
    cp1251_bulgarian_ci = 14
    latin1_danish_ci = 15
    hebrew_general_ci = 16
    tis620_thai_ci = 18
    euckr_korean_ci = 19
    latin7_estonian_cs = 20
    latin2_hungarian_ci = 21
    koi8u_general_ci = 22
    cp1251_ukrainian_ci = 23
    gb2312_chinese_ci = 24
    greek_general_ci = 25
    cp1250_general_ci = 26
    latin2_croatian_ci = 27
    gbk_chinese_ci = 28
    cp1257_lithuanian_ci = 29
    latin5_turkish_ci = 30
    latin1_german2_ci = 31
    armscii8_general_ci = 32
    utf8_general_ci = 33
    cp1250_czech_cs = 34
    ucs2_general_ci = 35
    cp866_general_ci = 36
    keybcs2_general_ci = 37
    macce_general_ci = 38
    macroman_general_ci = 39
    cp852_general_ci = 40
    latin7_general_ci = 41
    latin7_general_cs = 42
    macce_bin = 43
    cp1250_croatian_ci = 44
    utf8mb4_general_ci = 45
    utf8mb4_bin = 46
    latin1_bin = 47
    latin1_general_ci = 48
    latin1_general_cs = 49
    cp1251_bin = 50
    cp1251_general_ci = 51
    cp1251_general_cs = 52
    macroman_bin = 53
    utf16_general_ci = 54
    utf16_bin = 55
    utf16le_general_ci = 56
    cp1256_general_ci = 57
    cp1257_bin = 58
    cp1257_general_ci = 59
    utf32_general_ci = 60
    utf32_bin = 61
    utf16le_bin = 62
    binary = 63
    armscii8_bin = 64
    ascii_bin = 65
    cp1250_bin = 66
    cp1256_bin = 67
    cp866_bin = 68
    dec8_bin = 69
    greek_bin = 70
    hebrew_bin = 71
    hp8_bin = 72
    keybcs2_bin = 73
    koi8r_bin = 74
    koi8u_bin = 75
    latin2_bin = 77
    latin5_bin = 78
    latin7_bin = 79
    cp850_bin = 80
    cp852_bin = 81
    swe7_bin = 82
    utf8_bin = 83
    big5_bin = 84
    euckr_bin = 85
    gb2312_bin = 86
    gbk_bin = 87
    sjis_bin = 88
    tis620_bin = 89
    ucs2_bin = 90
    ujis_bin = 91
    geostd8_general_ci = 92
    geostd8_bin = 93
    latin1_spanish_ci = 94
    cp932_japanese_ci = 95
    cp932_bin = 96
    eucjpms_japanese_ci = 97
    eucjpms_bin = 98
    cp1250_polish_ci = 99
    utf16_unicode_ci = 101
    utf16_icelandic_ci = 102
    utf16_latvian_ci = 103
    utf16_romanian_ci = 104
    utf16_slovenian_ci = 105
    utf16_polish_ci = 106
    utf16_estonian_ci = 107
    utf16_spanish_ci = 108
    utf16_swedish_ci = 109
    utf16_turkish_ci = 110
    utf16_czech_ci = 111
    utf16_danish_ci = 112
    utf16_lithuanian_ci = 113
    utf16_slovak_ci = 114
    utf16_spanish2_ci = 115
    utf16_roman_ci = 116
    utf16_persian_ci = 117
    utf16_esperanto_ci = 118
    utf16_hungarian_ci = 119
    utf16_sinhala_ci = 120
    utf16_german2_ci = 121
    utf16_croatian_ci = 122
    utf16_unicode_520_ci = 123
    utf16_vietnamese_ci = 124
    ucs2_unicode_ci = 128
    ucs2_icelandic_ci = 129
    ucs2_latvian_ci = 130
    ucs2_romanian_ci = 131
    ucs2_slovenian_ci = 132
    ucs2_polish_ci = 133
    ucs2_estonian_ci = 134
    ucs2_spanish_ci = 135
    ucs2_swedish_ci = 136
    ucs2_turkish_ci = 137
    ucs2_czech_ci = 138
    ucs2_danish_ci = 139
    ucs2_lithuanian_ci = 140
    ucs2_slovak_ci = 141
    ucs2_spanish2_ci = 142
    ucs2_roman_ci = 143
    ucs2_persian_ci = 144
    ucs2_esperanto_ci = 145
    ucs2_hungarian_ci = 146
    ucs2_sinhala_ci = 147
    ucs2_german2_ci = 148
    ucs2_croatian_ci = 149
    ucs2_unicode_520_ci = 150
    ucs2_vietnamese_ci = 151
    ucs2_general_mysql500_ci = 159
    utf32_unicode_ci = 160
    utf32_icelandic_ci = 161
    utf32_latvian_ci = 162
    utf32_romanian_ci = 163
    utf32_slovenian_ci = 164
    utf32_polish_ci = 165
    utf32_estonian_ci = 166
    utf32_spanish_ci = 167
    utf32_swedish_ci = 168
    utf32_turkish_ci = 169
    utf32_czech_ci = 170
    utf32_danish_ci = 171
    utf32_lithuanian_ci = 172
    utf32_slovak_ci = 173
    utf32_spanish2_ci = 174
    utf32_roman_ci = 175
    utf32_persian_ci = 176
    utf32_esperanto_ci = 177
    utf32_hungarian_ci = 178
    utf32_sinhala_ci = 179
    utf32_german2_ci = 180
    utf32_croatian_ci = 181
    utf32_unicode_520_ci = 182
    utf32_vietnamese_ci = 183
    utf8_unicode_ci = 192
    utf8_icelandic_ci = 193
    utf8_latvian_ci = 194
    utf8_romanian_ci = 195
    utf8_slovenian_ci = 196
    utf8_polish_ci = 197
    utf8_estonian_ci = 198
    utf8_spanish_ci = 199
    utf8_swedish_ci = 200
    utf8_turkish_ci = 201
    utf8_czech_ci = 202
    utf8_danish_ci = 203
    utf8_lithuanian_ci = 204
    utf8_slovak_ci = 205
    utf8_spanish2_ci = 206
    utf8_roman_ci = 207
    utf8_persian_ci = 208
    utf8_esperanto_ci = 209
    utf8_hungarian_ci = 210
    utf8_sinhala_ci = 211
    utf8_german2_ci = 212
    utf8_croatian_ci = 213
    utf8_unicode_520_ci = 214
    utf8_vietnamese_ci = 215
    utf8_general_mysql500_ci = 223
    utf8mb4_unicode_ci = 224
    utf8mb4_icelandic_ci = 225
    utf8mb4_latvian_ci = 226
    utf8mb4_romanian_ci = 227
    utf8mb4_slovenian_ci = 228
    utf8mb4_polish_ci = 229
    utf8mb4_estonian_ci = 230
    utf8mb4_spanish_ci = 231
    utf8mb4_swedish_ci = 232
    utf8mb4_turkish_ci = 233
    utf8mb4_czech_ci = 234
    utf8mb4_danish_ci = 235
    utf8mb4_lithuanian_ci = 236
    utf8mb4_slovak_ci = 237
    utf8mb4_spanish2_ci = 238
    utf8mb4_roman_ci = 239
    utf8mb4_persian_ci = 240
    utf8mb4_esperanto_ci = 241
    utf8mb4_hungarian_ci = 242
    utf8mb4_sinhala_ci = 243
    utf8mb4_german2_ci = 244
    utf8mb4_croatian_ci = 245
    utf8mb4_unicode_520_ci = 246
    utf8mb4_vietnamese_ci = 247
    gb18030_chinese_ci = 248
    gb18030_bin = 249
    gb18030_unicode_520_ci = 250
    utf8mb4_0900_ai_ci = 255

    @property
    def codec(self) -> str:
        return self.charset.codec

    @property
    def charset(self) -> CharacterSet:
        return DEFAULT_CHARACTER_SETS[self]


DEFAULT_CHARACTER_SETS = {
    Collation.big5_chinese_ci: CharacterSet.big5,
    Collation.latin2_czech_cs: CharacterSet.latin2,
    Collation.dec8_swedish_ci: CharacterSet.dec8,
    Collation.cp850_general_ci: CharacterSet.cp850,
    Collation.latin1_german1_ci: CharacterSet.latin1,
    Collation.hp8_english_ci: CharacterSet.hp8,
    Collation.koi8r_general_ci: CharacterSet.koi8r,
    Collation.latin1_swedish_ci: CharacterSet.latin1,
    Collation.latin2_general_ci: CharacterSet.latin2,
    Collation.swe7_swedish_ci: CharacterSet.swe7,
    Collation.ascii_general_ci: CharacterSet.ascii,
    Collation.ujis_japanese_ci: CharacterSet.ujis,
    Collation.sjis_japanese_ci: CharacterSet.sjis,
    Collation.cp1251_bulgarian_ci: CharacterSet.cp1251,
    Collation.latin1_danish_ci: CharacterSet.latin1,
    Collation.hebrew_general_ci: CharacterSet.hebrew,
    Collation.tis620_thai_ci: CharacterSet.tis620,
    Collation.euckr_korean_ci: CharacterSet.euckr,
    Collation.latin7_estonian_cs: CharacterSet.latin7,
    Collation.latin2_hungarian_ci: CharacterSet.latin2,
    Collation.koi8u_general_ci: CharacterSet.koi8u,
    Collation.cp1251_ukrainian_ci: CharacterSet.cp1251,
    Collation.gb2312_chinese_ci: CharacterSet.gb2312,
    Collation.greek_general_ci: CharacterSet.greek,
    Collation.cp1250_general_ci: CharacterSet.cp1250,
    Collation.latin2_croatian_ci: CharacterSet.latin2,
    Collation.gbk_chinese_ci: CharacterSet.gbk,
    Collation.cp1257_lithuanian_ci: CharacterSet.cp1257,
    Collation.latin5_turkish_ci: CharacterSet.latin5,
    Collation.latin1_german2_ci: CharacterSet.latin1,
    Collation.armscii8_general_ci: CharacterSet.armscii8,
    Collation.utf8_general_ci: CharacterSet.utf8,
    Collation.cp1250_czech_cs: CharacterSet.cp1250,
    Collation.ucs2_general_ci: CharacterSet.ucs2,
    Collation.cp866_general_ci: CharacterSet.cp866,
    Collation.keybcs2_general_ci: CharacterSet.keybcs2,
    Collation.macce_general_ci: CharacterSet.macce,
    Collation.macroman_general_ci: CharacterSet.macroman,
    Collation.cp852_general_ci: CharacterSet.cp852,
    Collation.latin7_general_ci: CharacterSet.latin7,
    Collation.latin7_general_cs: CharacterSet.latin7,
    Collation.macce_bin: CharacterSet.macce,
    Collation.cp1250_croatian_ci: CharacterSet.cp1250,
    Collation.utf8mb4_general_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_bin: CharacterSet.utf8mb4,
    Collation.latin1_bin: CharacterSet.latin1,
    Collation.latin1_general_ci: CharacterSet.latin1,
    Collation.latin1_general_cs: CharacterSet.latin1,
    Collation.cp1251_bin: CharacterSet.cp1251,
    Collation.cp1251_general_ci: CharacterSet.cp1251,
    Collation.cp1251_general_cs: CharacterSet.cp1251,
    Collation.macroman_bin: CharacterSet.macroman,
    Collation.utf16_general_ci: CharacterSet.utf16,
    Collation.utf16_bin: CharacterSet.utf16,
    Collation.utf16le_general_ci: CharacterSet.utf16le,
    Collation.cp1256_general_ci: CharacterSet.cp1256,
    Collation.cp1257_bin: CharacterSet.cp1257,
    Collation.cp1257_general_ci: CharacterSet.cp1257,
    Collation.utf32_general_ci: CharacterSet.utf32,
    Collation.utf32_bin: CharacterSet.utf32,
    Collation.utf16le_bin: CharacterSet.utf16le,
    Collation.binary: CharacterSet.binary,
    Collation.armscii8_bin: CharacterSet.armscii8,
    Collation.ascii_bin: CharacterSet.ascii,
    Collation.cp1250_bin: CharacterSet.cp1250,
    Collation.cp1256_bin: CharacterSet.cp1256,
    Collation.cp866_bin: CharacterSet.cp866,
    Collation.dec8_bin: CharacterSet.dec8,
    Collation.greek_bin: CharacterSet.greek,
    Collation.hebrew_bin: CharacterSet.hebrew,
    Collation.hp8_bin: CharacterSet.hp8,
    Collation.keybcs2_bin: CharacterSet.keybcs2,
    Collation.koi8r_bin: CharacterSet.koi8r,
    Collation.koi8u_bin: CharacterSet.koi8u,
    Collation.latin2_bin: CharacterSet.latin2,
    Collation.latin5_bin: CharacterSet.latin5,
    Collation.latin7_bin: CharacterSet.latin7,
    Collation.cp850_bin: CharacterSet.cp850,
    Collation.cp852_bin: CharacterSet.cp852,
    Collation.swe7_bin: CharacterSet.swe7,
    Collation.utf8_bin: CharacterSet.utf8,
    Collation.big5_bin: CharacterSet.big5,
    Collation.euckr_bin: CharacterSet.euckr,
    Collation.gb2312_bin: CharacterSet.gb2312,
    Collation.gbk_bin: CharacterSet.gbk,
    Collation.sjis_bin: CharacterSet.sjis,
    Collation.tis620_bin: CharacterSet.tis620,
    Collation.ucs2_bin: CharacterSet.ucs2,
    Collation.ujis_bin: CharacterSet.ujis,
    Collation.geostd8_general_ci: CharacterSet.geostd8,
    Collation.geostd8_bin: CharacterSet.geostd8,
    Collation.latin1_spanish_ci: CharacterSet.latin1,
    Collation.cp932_japanese_ci: CharacterSet.cp932,
    Collation.cp932_bin: CharacterSet.cp932,
    Collation.eucjpms_japanese_ci: CharacterSet.eucjpms,
    Collation.eucjpms_bin: CharacterSet.eucjpms,
    Collation.cp1250_polish_ci: CharacterSet.cp1250,
    Collation.utf16_unicode_ci: CharacterSet.utf16,
    Collation.utf16_icelandic_ci: CharacterSet.utf16,
    Collation.utf16_latvian_ci: CharacterSet.utf16,
    Collation.utf16_romanian_ci: CharacterSet.utf16,
    Collation.utf16_slovenian_ci: CharacterSet.utf16,
    Collation.utf16_polish_ci: CharacterSet.utf16,
    Collation.utf16_estonian_ci: CharacterSet.utf16,
    Collation.utf16_spanish_ci: CharacterSet.utf16,
    Collation.utf16_swedish_ci: CharacterSet.utf16,
    Collation.utf16_turkish_ci: CharacterSet.utf16,
    Collation.utf16_czech_ci: CharacterSet.utf16,
    Collation.utf16_danish_ci: CharacterSet.utf16,
    Collation.utf16_lithuanian_ci: CharacterSet.utf16,
    Collation.utf16_slovak_ci: CharacterSet.utf16,
    Collation.utf16_spanish2_ci: CharacterSet.utf16,
    Collation.utf16_roman_ci: CharacterSet.utf16,
    Collation.utf16_persian_ci: CharacterSet.utf16,
    Collation.utf16_esperanto_ci: CharacterSet.utf16,
    Collation.utf16_hungarian_ci: CharacterSet.utf16,
    Collation.utf16_sinhala_ci: CharacterSet.utf16,
    Collation.utf16_german2_ci: CharacterSet.utf16,
    Collation.utf16_croatian_ci: CharacterSet.utf16,
    Collation.utf16_unicode_520_ci: CharacterSet.utf16,
    Collation.utf16_vietnamese_ci: CharacterSet.utf16,
    Collation.ucs2_unicode_ci: CharacterSet.ucs2,
    Collation.ucs2_icelandic_ci: CharacterSet.ucs2,
    Collation.ucs2_latvian_ci: CharacterSet.ucs2,
    Collation.ucs2_romanian_ci: CharacterSet.ucs2,
    Collation.ucs2_slovenian_ci: CharacterSet.ucs2,
    Collation.ucs2_polish_ci: CharacterSet.ucs2,
    Collation.ucs2_estonian_ci: CharacterSet.ucs2,
    Collation.ucs2_spanish_ci: CharacterSet.ucs2,
    Collation.ucs2_swedish_ci: CharacterSet.ucs2,
    Collation.ucs2_turkish_ci: CharacterSet.ucs2,
    Collation.ucs2_czech_ci: CharacterSet.ucs2,
    Collation.ucs2_danish_ci: CharacterSet.ucs2,
    Collation.ucs2_lithuanian_ci: CharacterSet.ucs2,
    Collation.ucs2_slovak_ci: CharacterSet.ucs2,
    Collation.ucs2_spanish2_ci: CharacterSet.ucs2,
    Collation.ucs2_roman_ci: CharacterSet.ucs2,
    Collation.ucs2_persian_ci: CharacterSet.ucs2,
    Collation.ucs2_esperanto_ci: CharacterSet.ucs2,
    Collation.ucs2_hungarian_ci: CharacterSet.ucs2,
    Collation.ucs2_sinhala_ci: CharacterSet.ucs2,
    Collation.ucs2_german2_ci: CharacterSet.ucs2,
    Collation.ucs2_croatian_ci: CharacterSet.ucs2,
    Collation.ucs2_unicode_520_ci: CharacterSet.ucs2,
    Collation.ucs2_vietnamese_ci: CharacterSet.ucs2,
    Collation.ucs2_general_mysql500_ci: CharacterSet.ucs2,
    Collation.utf32_unicode_ci: CharacterSet.utf32,
    Collation.utf32_icelandic_ci: CharacterSet.utf32,
    Collation.utf32_latvian_ci: CharacterSet.utf32,
    Collation.utf32_romanian_ci: CharacterSet.utf32,
    Collation.utf32_slovenian_ci: CharacterSet.utf32,
    Collation.utf32_polish_ci: CharacterSet.utf32,
    Collation.utf32_estonian_ci: CharacterSet.utf32,
    Collation.utf32_spanish_ci: CharacterSet.utf32,
    Collation.utf32_swedish_ci: CharacterSet.utf32,
    Collation.utf32_turkish_ci: CharacterSet.utf32,
    Collation.utf32_czech_ci: CharacterSet.utf32,
    Collation.utf32_danish_ci: CharacterSet.utf32,
    Collation.utf32_lithuanian_ci: CharacterSet.utf32,
    Collation.utf32_slovak_ci: CharacterSet.utf32,
    Collation.utf32_spanish2_ci: CharacterSet.utf32,
    Collation.utf32_roman_ci: CharacterSet.utf32,
    Collation.utf32_persian_ci: CharacterSet.utf32,
    Collation.utf32_esperanto_ci: CharacterSet.utf32,
    Collation.utf32_hungarian_ci: CharacterSet.utf32,
    Collation.utf32_sinhala_ci: CharacterSet.utf32,
    Collation.utf32_german2_ci: CharacterSet.utf32,
    Collation.utf32_croatian_ci: CharacterSet.utf32,
    Collation.utf32_unicode_520_ci: CharacterSet.utf32,
    Collation.utf32_vietnamese_ci: CharacterSet.utf32,
    Collation.utf8_unicode_ci: CharacterSet.utf8,
    Collation.utf8_icelandic_ci: CharacterSet.utf8,
    Collation.utf8_latvian_ci: CharacterSet.utf8,
    Collation.utf8_romanian_ci: CharacterSet.utf8,
    Collation.utf8_slovenian_ci: CharacterSet.utf8,
    Collation.utf8_polish_ci: CharacterSet.utf8,
    Collation.utf8_estonian_ci: CharacterSet.utf8,
    Collation.utf8_spanish_ci: CharacterSet.utf8,
    Collation.utf8_swedish_ci: CharacterSet.utf8,
    Collation.utf8_turkish_ci: CharacterSet.utf8,
    Collation.utf8_czech_ci: CharacterSet.utf8,
    Collation.utf8_danish_ci: CharacterSet.utf8,
    Collation.utf8_lithuanian_ci: CharacterSet.utf8,
    Collation.utf8_slovak_ci: CharacterSet.utf8,
    Collation.utf8_spanish2_ci: CharacterSet.utf8,
    Collation.utf8_roman_ci: CharacterSet.utf8,
    Collation.utf8_persian_ci: CharacterSet.utf8,
    Collation.utf8_esperanto_ci: CharacterSet.utf8,
    Collation.utf8_hungarian_ci: CharacterSet.utf8,
    Collation.utf8_sinhala_ci: CharacterSet.utf8,
    Collation.utf8_german2_ci: CharacterSet.utf8,
    Collation.utf8_croatian_ci: CharacterSet.utf8,
    Collation.utf8_unicode_520_ci: CharacterSet.utf8,
    Collation.utf8_vietnamese_ci: CharacterSet.utf8,
    Collation.utf8_general_mysql500_ci: CharacterSet.utf8,
    Collation.utf8mb4_unicode_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_icelandic_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_latvian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_romanian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_slovenian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_polish_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_estonian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_spanish_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_swedish_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_turkish_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_czech_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_danish_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_lithuanian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_slovak_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_spanish2_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_roman_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_persian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_esperanto_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_hungarian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_sinhala_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_german2_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_croatian_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_unicode_520_ci: CharacterSet.utf8mb4,
    Collation.utf8mb4_vietnamese_ci: CharacterSet.utf8mb4,
    Collation.gb18030_chinese_ci: CharacterSet.gb18030,
    Collation.gb18030_bin: CharacterSet.gb18030,
    Collation.gb18030_unicode_520_ci: CharacterSet.gb18030,
    Collation.utf8mb4_0900_ai_ci: CharacterSet.utf8mb4,
}
DEFAULT_COLLATIONS = {
    CharacterSet.big5: Collation.big5_chinese_ci,
    CharacterSet.dec8: Collation.dec8_swedish_ci,
    CharacterSet.cp850: Collation.cp850_general_ci,
    CharacterSet.hp8: Collation.hp8_english_ci,
    CharacterSet.koi8r: Collation.koi8r_general_ci,
    CharacterSet.latin1: Collation.latin1_swedish_ci,
    CharacterSet.latin2: Collation.latin2_general_ci,
    CharacterSet.swe7: Collation.swe7_swedish_ci,
    CharacterSet.ascii: Collation.ascii_general_ci,
    CharacterSet.ujis: Collation.ujis_japanese_ci,
    CharacterSet.sjis: Collation.sjis_japanese_ci,
    CharacterSet.hebrew: Collation.hebrew_general_ci,
    CharacterSet.tis620: Collation.tis620_thai_ci,
    CharacterSet.euckr: Collation.euckr_korean_ci,
    CharacterSet.koi8u: Collation.koi8u_general_ci,
    CharacterSet.gb2312: Collation.gb2312_chinese_ci,
    CharacterSet.greek: Collation.greek_general_ci,
    CharacterSet.cp1250: Collation.cp1250_general_ci,
    CharacterSet.gbk: Collation.gbk_chinese_ci,
    CharacterSet.latin5: Collation.latin5_turkish_ci,
    CharacterSet.armscii8: Collation.armscii8_general_ci,
    CharacterSet.utf8: Collation.utf8_general_ci,
    CharacterSet.ucs2: Collation.ucs2_general_ci,
    CharacterSet.cp866: Collation.cp866_general_ci,
    CharacterSet.keybcs2: Collation.keybcs2_general_ci,
    CharacterSet.macce: Collation.macce_general_ci,
    CharacterSet.macroman: Collation.macroman_general_ci,
    CharacterSet.cp852: Collation.cp852_general_ci,
    CharacterSet.latin7: Collation.latin7_general_ci,
    CharacterSet.utf8mb4: Collation.utf8mb4_general_ci,
    CharacterSet.cp1251: Collation.cp1251_general_ci,
    CharacterSet.utf16: Collation.utf16_general_ci,
    CharacterSet.utf16le: Collation.utf16le_general_ci,
    CharacterSet.cp1256: Collation.cp1256_general_ci,
    CharacterSet.cp1257: Collation.cp1257_general_ci,
    CharacterSet.utf32: Collation.utf32_general_ci,
    CharacterSet.binary: Collation.binary,
    CharacterSet.geostd8: Collation.geostd8_general_ci,
    CharacterSet.cp932: Collation.cp932_japanese_ci,
    CharacterSet.eucjpms: Collation.eucjpms_japanese_ci,
    CharacterSet.gb18030: Collation.gb18030_chinese_ci,
}
